@@ -1,3 +1,4 @@
+#include <sys/queue.h>
 #include <alloca.h>
 #include <ctype.h>
 #include <dirent.h>
@@ -29,30 +30,34 @@ int selinux_enforce = 0;
 #define GLOB_ABORTED GLOB_ABEND
 #endif
 
-typedef struct {
+struct logState {
     char *fn;
     struct tm lastRotated;	/* only tm.mon, tm_mday, tm_year are good! */
     struct stat sb;
     int doRotate;
-} logState;
+    LIST_ENTRY(logState) list;
+};
 
-typedef struct {
+struct logNames {
     char *firstRotated;
     char *disposeName;
     char *finalName;
     char *dirName;
     char *baseName;
-} logNames;
-
-struct stateSet {
-    logState *states;
-    int numStates;
 };
 
+struct logStates {
+	LIST_HEAD(stateSet, logState) head;
+} **states;
+
+unsigned int hashSize;
+int numLogs = 0;
 int debug = 0;
 char *mailCommand = DEFAULT_MAIL_COMMAND;
 time_t nowSecs = 0;
 
+static int shred_file(char *filename, struct logInfo *log);
+
 static int globerr(const char *pathname, int theerr)
 {
     message(MESS_ERROR, "error accessing %s: %s\n", pathname,
@@ -62,80 +67,103 @@ static int globerr(const char *pathname, int theerr)
     return 1;
 }
 
-/* logInfo instances may share dynamic allocated memory, so to avoid
- * free memory already free'd, it is necessary check if a item in a
- * logInfo instance points to the same memory address of the item in
- * previous instances, which were free'd before. */
-
-#define freeLogItem(what) \
-    if (log->what) { \
-        for (j = 0; j < i; j++) \
-            if (log->what == (*logsPtr + j)->what) \
-                break; \
-        if (j == i) \
-            free(log->what); \
-    }
-
-static void free_logInfo(logInfo ** logsPtr, int *numLogsPtr)
+#define HASH_SIZE_MIN 64
+static int allocateHash(void)
 {
-    int i, j;
-    logInfo *log;
+	struct logInfo *log;
+	unsigned int hs;
+	int i;
 
-    for (i = 0; i < *numLogsPtr; i++) {
-	log = *logsPtr + i;
-	freeLogItem(pattern);
-	freeLogItem(oldDir);
-	freeLogItem(pre);
-	freeLogItem(post);
-	freeLogItem(first);
-	freeLogItem(last);
-	freeLogItem(logAddress);
-	freeLogItem(extension);
-	freeLogItem(compress_prog);
-	freeLogItem(uncompress_prog);
-	freeLogItem(compress_ext);
-	freeLogItem(compress_options_list);
+	hs = 0;
 
-	for (j = 0; j < log->numFiles; j++)
-	    free(log->files[j]);
-	free(log->files);
-    }
-    free(*logsPtr);
-    *numLogsPtr = 0;
+	for (log = logs.tqh_first; log != NULL; log = log->list.tqe_next)
+		hs += log->numFiles;
+
+	hs *= 2;
+
+	/* Enforce some reasonable minimum hash size */
+	if (hs < HASH_SIZE_MIN)
+		hs = HASH_SIZE_MIN;
+
+	states = calloc(hs, sizeof(struct logStates *));
+	if (states == NULL) {
+		message(MESS_ERROR, "could not allocate memory for "
+				"hash table\n");
+		return 1;
+	}
+
+	for (i = 0; i < hs; i++) {
+		states[i] = malloc(sizeof(struct logState));
+		if (states[i] == NULL) {
+			message(MESS_ERROR, "could not allocate memory for "
+				"hash element\n");
+			return 1;
+		}
+		LIST_INIT(&(states[i]->head));
+	}
+
+	hashSize = hs;
+
+	return 0;
 }
 
-static logState *findState(const char *fn, struct stateSet *sip)
+#define HASH_CONST 13
+static unsigned hashIndex(const char *fn)
 {
-    int i;
-    logState *states = sip->states;
-    int numStates = sip->numStates;
-    struct tm now = *localtime(&nowSecs);
-    time_t lr_time;
+	unsigned hash = 0;
 
-    for (i = 0; i < numStates; i++)
-	if (!strcmp(fn, states[i].fn))
-	    break;
+	while (*fn) {
+		hash *= HASH_CONST;
+		hash += *fn++;
+	}
 
-    if (i == numStates) {
-	i = numStates++;
-	states = realloc(states, sizeof(*states) * numStates);
-	states[i].fn = strdup(fn);
-	memset(&states[i].lastRotated, 0, sizeof(states[i].lastRotated));
-	states[i].doRotate = 0;
+	return hash % hashSize;
+}
 
-	states[i].lastRotated.tm_mon = now.tm_mon;
-	states[i].lastRotated.tm_mday = now.tm_mday;
-	states[i].lastRotated.tm_year = now.tm_year;
+static struct logState *newState(const char *fn)
+{
+	struct tm now = *localtime(&nowSecs);
+	struct logState *new;
+	time_t lr_time;
 
-	/* fill in the rest of the st->lastRotated fields */
-	lr_time = mktime(&states[i].lastRotated);
-	states[i].lastRotated = *localtime(&lr_time);
+	if ((new = malloc(sizeof(*new))) == NULL)
+		return NULL;
 
-	sip->states = states;
-	sip->numStates = numStates;
-    }
+	if ((new->fn = strdup(fn)) == NULL)
+		return NULL;
 
-    return (states + i);
+	new->doRotate = 0;
+
+	memset(&new->lastRotated, 0, sizeof(new->lastRotated));
+	new->lastRotated.tm_mon = now.tm_mon;
+	new->lastRotated.tm_mday = now.tm_mday;
+	new->lastRotated.tm_year = now.tm_year;
+
+	/* fill in the rest of the new->lastRotated fields */
+	lr_time = mktime(&new->lastRotated);
+	new->lastRotated = *localtime(&lr_time);
+
+	return new;
+}
+
+static struct logState *findState(const char *fn)
+{
+	unsigned int i = hashIndex(fn);
+	struct logState *p;
+
+	for (p = states[i]->head.lh_first; p != NULL; p = p->list.le_next)
+		if (!strcmp(fn, p->fn))
+			break;
+
+	/* new state */
+	if (p == NULL) {
+		if ((p = newState(fn)) == NULL)
+			return NULL;
+
+		LIST_INSERT_HEAD(&(states[i]->head), p, list);
+	}
+
+	return p;
 }
 
 static int runScript(char *logfn, char *script)
@@ -148,10 +176,10 @@ static int runScript(char *logfn, char *script)
 	return 0;
     }
 
-    if (!fork()) {
-	execl("/bin/sh", "sh", "-c", script, NULL);
-	exit(1);
-    }
+	if (!fork()) {
+		execl("/bin/sh", "sh", "-c", script, "logrotate_script", logfn, NULL);
+		exit(1);
+	}
 
     wait(&rc);
     return rc;
@@ -188,11 +216,54 @@ int createOutputFile(char *fileName, int flags, struct stat *sb)
     return fd;
 }
 
-static int removeLogFile(char *name)
+#define SHRED_CALL "shred -u "
+#define SHRED_COUNT_FLAG "-n "
+#define DIGITS 10
+/* unlink, but try to call shred from GNU fileutils */
+static int shred_file(char *filename, struct logInfo *log)
+{
+	int len, ret;
+	char *cmd;
+	char count[DIGITS];    /*  that's a lot of shredding :)  */
+
+	if (!(log->flags & LOG_FLAG_SHRED)) {
+		return unlink(filename);
+	}
+
+	len = strlen(filename) + strlen(SHRED_CALL);
+	len += strlen(SHRED_COUNT_FLAG) + DIGITS;
+	cmd = malloc(len);
+
+	if (!cmd) {
+		message(MESS_ERROR, "malloc error while shredding");
+		return unlink(filename);
+	}
+	strcpy(cmd, SHRED_CALL);
+	if (log->shred_cycles != 0) {
+		strcat(cmd, SHRED_COUNT_FLAG);
+		snprintf(count, DIGITS - 1, "%d", log->shred_cycles);
+		strcat(count, " ");
+		strcat(cmd, count);
+	}
+	strcat(cmd, filename);
+	ret = system(cmd);
+	free(cmd);
+	if (ret != 0) {
+		message(MESS_ERROR, "Failed to shred %s\n, trying unlink", filename);
+		if (ret != -1) {
+			message(MESS_NORMAL, "Shred returned %d\n", ret);
+		}
+		return unlink(filename);
+	} else {
+		return ret;
+	}
+}
+
+static int removeLogFile(char *name, struct logInfo *log)
 {
     message(MESS_DEBUG, "removing old log %s\n", name);
 
-    if (!debug && unlink(name)) {
+    if (!debug && shred_file(name, log)) {
 	message(MESS_ERROR, "Failed to remove old log %s: %s\n",
 		name, strerror(errno));
 	return 1;
@@ -200,7 +271,7 @@ static int removeLogFile(char *name)
     return 0;
 }
 
-static int compressLogFile(char *name, logInfo * log, struct stat *sb)
+static int compressLogFile(char *name, struct logInfo *log, struct stat *sb)
 {
     char *compressedName;
     const char **fullCommand;
@@ -255,7 +326,7 @@ static int compressLogFile(char *name, logInfo * log, struct stat *sb)
 	return 1;
     }
 
-    unlink(name);
+    shred_file(name, log);
 
     return 0;
 }
@@ -264,7 +335,7 @@ static int mailLog(char *logFile, char *mailCommand,
 		   char *uncompressCommand, char *address, char *subject)
 {
     int mailInput;
-    pid_t mailChild, uncompressChild;
+    pid_t mailChild, uncompressChild = 0;
     int mailStatus, uncompressStatus;
     int uncompressPipe[2];
     char *mailArgv[] = { mailCommand, "-s", subject, address, NULL };
@@ -277,22 +348,26 @@ static int mailLog(char *logFile, char *mailCommand,
     }
 
     if (uncompressCommand) {
-	pipe(uncompressPipe);
-	if (!(uncompressChild = fork())) {
-	    /* uncompress child */
-	    dup2(mailInput, 0);
-	    close(mailInput);
-	    dup2(uncompressPipe[1], 1);
-	    close(uncompressPipe[0]);
-	    close(uncompressPipe[1]);
+		if (pipe(uncompressPipe) < 0) {
+			message(MESS_ERROR, "error opening pipe for uncompress: %s",
+					strerror(errno));
+			return 1;
+		}
+		if (!(uncompressChild = fork())) {
+			/* uncompress child */
+			dup2(mailInput, 0);
+			close(mailInput);
+			dup2(uncompressPipe[1], 1);
+			close(uncompressPipe[0]);
+			close(uncompressPipe[1]);
 
-	    execlp(uncompressCommand, uncompressCommand, NULL);
-	    exit(1);
-	}
+			execlp(uncompressCommand, uncompressCommand, NULL);
+			exit(1);
+		}
 
-	close(mailInput);
-	mailInput = uncompressPipe[0];
-	close(uncompressPipe[1]);
+		close(mailInput);
+		mailInput = uncompressPipe[0];
+		close(uncompressPipe[1]);
     }
 
     if (!(mailChild = fork())) {
@@ -327,7 +402,7 @@ static int mailLog(char *logFile, char *mailCommand,
 }
 
 static int mailLogWrapper(char *mailFilename, char *mailCommand,
-			  int logNum, logInfo * log)
+			  int logNum, struct logInfo *log)
 {
     /* if the log is compressed (and we're not mailing a
      * file whose compression has been delayed), we need
@@ -371,19 +446,20 @@ static int copyTruncate(char *currLog, char *saveLog, struct stat *sb,
 			    "getting default context: %s\n",
 			    strerror(errno));
 		    if (selinux_enforce) {
-			freecon(oldContext);
-			return 1;
+				freecon(oldContext);
+				return 1;
 		    }
 		}
 		if (setfscreatecon_raw(oldContext) < 0) {
 		    message(MESS_ERROR,
 			    "setting file context %s to %s: %s\n",
 			    saveLog, oldContext, strerror(errno));
-		    if (selinux_enforce) {
-			freecon(oldContext);
-			return 1;
+			if (selinux_enforce) {
+				freecon(oldContext);
+				return 1;
 		    }
 		}
+		message(MESS_DEBUG, "set default create context\n");
 		freecon(oldContext);
 	    } else {
 		    if (errno != ENOTSUP) {
@@ -401,10 +477,8 @@ static int copyTruncate(char *currLog, char *saveLog, struct stat *sb,
 #ifdef WITH_SELINUX
 	if (selinux_enabled) {
 	    setfscreatecon_raw(prev_context);
-	    if (prev_context != NULL) {
 		freecon(prev_context);
 		prev_context = NULL;
-	    }
 	}
 #endif
 	if (fdsave < 0) {
@@ -448,10 +522,10 @@ static int copyTruncate(char *currLog, char *saveLog, struct stat *sb,
     return 0;
 }
 
-int findNeedRotating(logInfo * log, int logNum, struct stateSet *sip)
+int findNeedRotating(struct logInfo *log, int logNum)
 {
     struct stat sb;
-    logState *state = NULL;
+    struct logState *state = NULL;
     struct tm now = *localtime(&nowSecs);
 
     message(MESS_DEBUG, "considering log %s\n", log->files[logNum]);
@@ -467,7 +541,7 @@ int findNeedRotating(logInfo * log, int logNum, struct stateSet *sip)
 	return 1;
     }
 
-    state = findState(log->files[logNum], sip);
+    state = findState(log->files[logNum]);
     state->doRotate = 0;
     state->sb = sb;
 
@@ -538,8 +612,8 @@ int findNeedRotating(logInfo * log, int logNum, struct stateSet *sip)
     return 0;
 }
 
-int prerotateSingleLog(logInfo * log, int logNum, logState * state,
-		       logNames * rotNames)
+int prerotateSingleLog(struct logInfo *log, int logNum, struct logState *state,
+		       struct logNames *rotNames)
 {
     struct tm now = *localtime(&nowSecs);
     char *oldName, *newName = NULL;
@@ -547,13 +621,18 @@ int prerotateSingleLog(logInfo * log, int logNum, logState * state,
     char *compext = "";
     char *fileext = "";
     int hasErrors = 0;
-    int i;
+    int i, j;
     char *glob_pattern;
     glob_t globResult;
     int rc;
-    size_t alloc_size;
     int rotateCount = log->rotateCount ? log->rotateCount : 1;
     int logStart = (log->logStart == -1) ? 1 : log->logStart;
+#define DATEEXT_LEN 64
+#define PATTERN_LEN (DATEEXT_LEN * 2)
+	char dext_str[DATEEXT_LEN];
+	char dformat[DATEEXT_LEN];
+	char dext_pattern[PATTERN_LEN];
+	char *dext;
 
     if (!state->doRotate)
 	return 0;
@@ -583,13 +662,9 @@ int prerotateSingleLog(logInfo * log, int logNum, logState * state,
 
     rotNames->baseName = strdup(ourBaseName(log->files[logNum]));
 
-    alloc_size = strlen(rotNames->dirName) + strlen(rotNames->baseName) +
-	strlen(log->files[logNum]) + strlen(fileext) +
-	strlen(compext) + 18;
-
-    oldName = alloca(alloc_size);
-    newName = alloca(alloc_size);
-    rotNames->disposeName = malloc(alloc_size);
+    oldName = alloca(PATH_MAX);
+    newName = alloca(PATH_MAX);
+    rotNames->disposeName = malloc(PATH_MAX);
 
     if (log->extension &&
 	strncmp(&
@@ -608,26 +683,101 @@ int prerotateSingleLog(logInfo * log, int logNum, logState * state,
 	free(rotNames->baseName);
 	rotNames->baseName = tempstr;
     }
+	
+	/* Allow only %Y %d %m and create valid strftime format string
+	 * Construct the glob pattern corresponding to the date format */
+	dext_str[0] = '\0';
+	if (log->dateformat) {
+		i = j = 0;
+		memset(dext_pattern, 0, sizeof(dext_pattern));
+		dext = log->dateformat;
+		while (*dext == ' ')
+			dext++;
+		while ((*dext != '\0') && (!hasErrors)) {
+			/* Will there be a space for a char and '\0'? */
+			if (j >= (sizeof(dext_pattern) - 1)) {
+				message(MESS_ERROR, "Date format %s is too long\n",
+						log->dateformat);
+				hasErrors = 1;
+				break;
+			}
+			if (*dext == '%') {
+				switch (*(dext + 1)) {
+					case 'Y':
+						strncat(dext_pattern, "[0-9][0-9]",
+								sizeof(dext_pattern) - strlen(dext_pattern));
+						j += 10; /* strlen("[0-9][0-9]") */
+					case 'm':
+					case 'd':
+						strncat(dext_pattern, "[0-9][0-9]",
+								sizeof(dext_pattern) - strlen(dext_pattern));
+						j += 10;
+						if (j >= (sizeof(dext_pattern) - 1)) {
+							message(MESS_ERROR, "Date format %s is too long\n",
+									log->dateformat);
+							hasErrors = 1;
+							break;
+						}
+						dformat[i++] = *(dext++);
+						dformat[i] = *dext;
+						break;
+					case 's':
+						/* End of year 2293 this pattern does not work. */
+						strncat(dext_pattern,
+								"[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9]",
+								sizeof(dext_pattern) - strlen(dext_pattern));
+						j += 50;
+						if (j >= (sizeof(dext_pattern) - 1)) {
+							message(MESS_ERROR, "Date format %s is too long\n",
+									log->dateformat);
+							hasErrors = 1;
+							break;
+						}
+						dformat[i++] = *(dext++);
+						dformat[i] = *dext;
+						break;
+					default:
+						dformat[i++] = *dext;
+						dformat[i] = '%';
+						dext_pattern[j++] = *dext;
+						break;
+				}
+			} else {
+				dformat[i] = *dext;
+				dext_pattern[j++] = *dext;
+			}
+			++i;
+			++dext;
+		}
+		dformat[i] = '\0';
+		message(MESS_DEBUG, "Converted '%s' -> '%s'\n", log->dateformat, dformat);
+		strftime(dext_str, sizeof(dext_str), dformat, &now);
+	} else {
+		/* The default dateformat and glob pattern */
+		strftime(dext_str, sizeof(dext_str), "-%Y%m%d", &now);
+		strncpy(dext_pattern, "-[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9]",
+				sizeof(dext_pattern));
+		dext_pattern[PATTERN_LEN - 1] = '\0';
+	}
+	message(MESS_DEBUG, "dateext suffix '%s'\n", dext_str);
+	message(MESS_DEBUG, "glob pattern '%s'\n", dext_pattern);
 
     /* First compress the previous log when necessary */
     if (log->flags & LOG_FLAG_COMPRESS &&
 	log->flags & LOG_FLAG_DELAYCOMPRESS) {
 	if (log->flags & LOG_FLAG_DATEEXT) {
-	    /* glob for uncompressed files with our pattern */
-	    glob_pattern =
-		malloc(strlen(rotNames->dirName) +
-		       strlen(rotNames->baseName)
-		       + strlen(fileext) + 44);
-	    sprintf(glob_pattern,
-		    "%s/%s-[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9]%s",
-		    rotNames->dirName, rotNames->baseName, fileext);
+		/* glob for uncompressed files with our pattern */
+		if (asprintf(&glob_pattern, "%s/%s%s%s", rotNames->dirName,
+					rotNames->baseName, dext_pattern, fileext) < 0) {
+			message(MESS_ERROR, "could not allocate glob pattern memory\n");
+		}
 	    rc = glob(glob_pattern, 0, globerr, &globResult);
 	    if (!rc && globResult.gl_pathc > 0) {
 		for (i = 0; i < globResult.gl_pathc && !hasErrors; i++) {
 		    struct stat sbprev;
 
-		    sprintf(oldName, "%s", (globResult.gl_pathv)[i]);
-		    if (stat(oldName, &sbprev)) {
+			snprintf(oldName, PATH_MAX, "%s", (globResult.gl_pathv)[i]);
+			if (stat(oldName, &sbprev)) {
 			message(MESS_DEBUG,
 				"previous log %s does not exist\n",
 				oldName);
@@ -639,7 +789,7 @@ int prerotateSingleLog(logInfo * log, int logNum, logState * state,
 		message(MESS_DEBUG,
 			"glob finding logs to compress failed\n");
 		/* fallback to old behaviour */
-		sprintf(oldName, "%s/%s.%d%s", rotNames->dirName,
+		snprintf(oldName, PATH_MAX, "%s/%s.%d%s", rotNames->dirName,
 			rotNames->baseName, logStart, fileext);
 	    }
 	    globfree(&globResult);
@@ -647,7 +797,7 @@ int prerotateSingleLog(logInfo * log, int logNum, logState * state,
 	} else {
 	    struct stat sbprev;
 
-	    sprintf(oldName, "%s/%s.%d%s", rotNames->dirName,
+	    snprintf(oldName, PATH_MAX, "%s/%s.%d%s", rotNames->dirName,
 		    rotNames->baseName, logStart, fileext);
 	    if (stat(oldName, &sbprev)) {
 		message(MESS_DEBUG, "previous log %s does not exist\n",
@@ -665,12 +815,10 @@ int prerotateSingleLog(logInfo * log, int logNum, logState * state,
     if (log->flags & LOG_FLAG_DATEEXT) {
 	/* glob for compressed files with our pattern
 	 * and compress ext */
-	glob_pattern =
-	    malloc(strlen(rotNames->dirName) + strlen(rotNames->baseName)
-		   + strlen(fileext) + strlen(compext) + 44);
-	sprintf(glob_pattern,
-		"%s/%s-[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9]%s%s",
-		rotNames->dirName, rotNames->baseName, fileext, compext);
+	if (asprintf(&glob_pattern, "%s/%s%s%s%s", rotNames->dirName,
+				rotNames->baseName, dext_pattern, fileext, compext) < 0) {
+		message(MESS_ERROR, "could not allocate glob pattern memory\n");
+	}
 	rc = glob(glob_pattern, 0, globerr, &globResult);
 	if (!rc) {
 	    /* search for files to drop, if we find one remember it,
@@ -697,7 +845,8 @@ int prerotateSingleLog(logInfo * log, int logNum, logState * state,
 						   mailCommand, logNum,
 						   log);
 			    if (!hasErrors)
-				hasErrors = removeLogFile(mailFilename);
+				message(MESS_DEBUG, "removing %s\n", mailFilename);
+				hasErrors = removeLogFile(mailFilename, log);
 			}
 			mail_out = i;
 		    }
@@ -705,7 +854,7 @@ int prerotateSingleLog(logInfo * log, int logNum, logState * state,
 	    }
 	    if (mail_out != -1) {
 		/* oldName is oldest Backup found (for unlink later) */
-		sprintf(oldName, "%s", (globResult.gl_pathv)[mail_out]);
+		snprintf(oldName, PATH_MAX, "%s", (globResult.gl_pathv)[mail_out]);
 		strcpy(rotNames->disposeName, oldName);
 	    } else {
 		free(rotNames->disposeName);
@@ -717,16 +866,15 @@ int prerotateSingleLog(logInfo * log, int logNum, logState * state,
 	    rotNames->disposeName = NULL;
 	}
 	/* firstRotated is most recently created/compressed rotated log */
-	sprintf(rotNames->firstRotated, "%s/%s-%04d%02d%02d%s%s",
-		rotNames->dirName, rotNames->baseName, now.tm_year + 1900,
-		now.tm_mon + 1, now.tm_mday, fileext, compext);
+	sprintf(rotNames->firstRotated, "%s/%s%s%s%s",
+		rotNames->dirName, rotNames->baseName, dext_str, fileext, compext);
 	globfree(&globResult);
 	free(glob_pattern);
     } else {
 	if (log->rotateAge) {
 	    struct stat fst_buf;
-	    for (i = 1; i <= rotateCount; i++) {
-		sprintf(oldName, "%s/%s.%d%s%s", rotNames->dirName,
+	    for (i = 1; i <= rotateCount + 1; i++) {
+		snprintf(oldName, PATH_MAX, "%s/%s.%d%s%s", rotNames->dirName,
 			rotNames->baseName, i, fileext, compext);
 		if (!stat(oldName, &fst_buf)
 		    && (((nowSecs - fst_buf.st_mtime) / 60 / 60 / 24)
@@ -737,12 +885,12 @@ int prerotateSingleLog(logInfo * log, int logNum, logState * state,
 			    mailLogWrapper(mailFilename, mailCommand,
 					   logNum, log);
 		    if (!hasErrors)
-			hasErrors = removeLogFile(mailFilename);
+			hasErrors = removeLogFile(mailFilename, log);
 		}
 	    }
 	}
 
-	sprintf(oldName, "%s/%s.%d%s%s", rotNames->dirName,
+	snprintf(oldName, PATH_MAX, "%s/%s.%d%s%s", rotNames->dirName,
 		rotNames->baseName, logStart + rotateCount, fileext,
 		compext);
 	strcpy(newName, oldName);
@@ -757,26 +905,26 @@ int prerotateSingleLog(logInfo * log, int logNum, logState * state,
 	if (selinux_enabled) {
 	    security_context_t oldContext = NULL;
 	    if (getfilecon_raw(log->files[logNum], &oldContext) > 0) {
-		if (getfscreatecon_raw(&prev_context) < 0) {
-		    message(MESS_ERROR,
-			    "getting default context: %s\n",
-			    strerror(errno));
-		    if (selinux_enforce) {
+			if (getfscreatecon_raw(&prev_context) < 0) {
+				message(MESS_ERROR,
+					"getting default context: %s\n",
+					strerror(errno));
+				if (selinux_enforce) {
+					freecon(oldContext);
+					return 1;
+				}
+			}
+			if (setfscreatecon_raw(oldContext) < 0) {
+				message(MESS_ERROR,
+					"setting file context %s to %s: %s\n",
+					log->files[logNum], oldContext,
+					strerror(errno));
+				if (selinux_enforce) {
+					freecon(oldContext);
+					return 1;
+				}
+			}
 			freecon(oldContext);
-			return 1;
-		    }
-		}
-		if (setfscreatecon_raw(oldContext) < 0) {
-		    message(MESS_ERROR,
-			    "setting file context %s to %s: %s\n",
-			    log->files[logNum], oldContext,
-			    strerror(errno));
-		    if (selinux_enforce) {
-			freecon(oldContext);
-			return 1;
-		    }
-		}
-		freecon(oldContext);
 	    } else {
 		if (errno != ENOENT && errno != ENOTSUP) {
 			message(MESS_ERROR, "getting file context %s: %s\n",
@@ -792,7 +940,7 @@ int prerotateSingleLog(logInfo * log, int logNum, logState * state,
 	    tmp = newName;
 	    newName = oldName;
 	    oldName = tmp;
-	    sprintf(oldName, "%s/%s.%d%s%s", rotNames->dirName,
+		snprintf(oldName, PATH_MAX, "%s/%s.%d%s%s", rotNames->dirName,
 		    rotNames->baseName, i, fileext, compext);
 
 	    message(MESS_DEBUG,
@@ -812,28 +960,28 @@ int prerotateSingleLog(logInfo * log, int logNum, logState * state,
 	}
     }				/* !LOG_FLAG_DATEEXT */
 
-    rotNames->finalName = malloc(alloc_size);
+	if (log->flags & LOG_FLAG_DATEEXT) {
+		char *destFile = alloca(PATH_MAX);
+		struct stat fst_buf;
 
-    if (log->flags & LOG_FLAG_DATEEXT) {
-	char *destFile =
-	    alloca(strlen(rotNames->dirName) + strlen(rotNames->baseName) +
-		   strlen(fileext) + strlen(compext) + 30);
-	struct stat fst_buf;
-	sprintf(rotNames->finalName, "%s/%s-%04d%02d%02d%s",
-		rotNames->dirName, rotNames->baseName, now.tm_year + 1900,
-		now.tm_mon + 1, now.tm_mday, fileext);
-	sprintf(destFile, "%s%s", rotNames->finalName, compext);
-	if (!stat(destFile, &fst_buf)) {
-	    message(MESS_DEBUG,
-		    "destination %s already exists, skipping rotation\n",
-		    rotNames->firstRotated);
-	    hasErrors = 1;
+		if (asprintf(&(rotNames->finalName), "%s/%s%s%s", rotNames->dirName,
+					rotNames->baseName, dext_str, fileext) < 0) {
+			message(MESS_ERROR, "could not allocate finalName memory\n");
+		}
+		snprintf(destFile, PATH_MAX, "%s%s", rotNames->finalName, compext);
+		if (!stat(destFile, &fst_buf)) {
+			message(MESS_DEBUG,
+					"destination %s already exists, skipping rotation\n",
+					rotNames->firstRotated);
+			hasErrors = 1;
+		}
+	} else {
+		/* note: the gzip extension is *not* used here! */
+		if (asprintf(&(rotNames->finalName), "%s/%s.%d%s", rotNames->dirName,
+					rotNames->baseName, logStart, fileext) < 0) {
+			message(MESS_ERROR, "could not allocate finalName memory\n");
+		}
 	}
-    } else {
-	/* note: the gzip extension is *not* used here! */
-	sprintf(rotNames->finalName, "%s/%s.%d%s", rotNames->dirName,
-		rotNames->baseName, logStart, fileext);
-    }
 
     /* if the last rotation doesn't exist, that's okay */
     if (!debug && rotNames->disposeName
@@ -848,12 +996,15 @@ int prerotateSingleLog(logInfo * log, int logNum, logState * state,
     return hasErrors;
 }
 
-int rotateSingleLog(logInfo * log, int logNum, logState * state,
-		    logNames * rotNames)
+int rotateSingleLog(struct logInfo *log, int logNum, struct logState *state,
+		    struct logNames *rotNames)
 {
     int hasErrors = 0;
     struct stat sb;
     int fd;
+#ifdef WITH_SELINUX
+	security_context_t savedContext = NULL;
+#endif
 
     if (!state->doRotate)
 	return 0;
@@ -861,9 +1012,64 @@ int rotateSingleLog(logInfo * log, int logNum, logState * state,
     if (!hasErrors) {
 
 	if (!(log->flags & (LOG_FLAG_COPYTRUNCATE | LOG_FLAG_COPY))) {
-	    message(MESS_DEBUG, "renaming %s to %s\n", log->files[logNum],
-		    rotNames->finalName);
+#ifdef WITH_SELINUX
+		if (selinux_enabled) {
+			security_context_t oldContext = NULL;
+			int fdcurr = -1;
 
+			if ((fdcurr = open(log->files[logNum], O_RDWR)) < 0) {
+				message(MESS_ERROR, "error opening %s: %s\n",
+						log->files[logNum],
+					strerror(errno));
+				return 1;
+			}
+			if (fgetfilecon_raw(fdcurr, &oldContext) >= 0) {
+				if (getfscreatecon_raw(&savedContext) < 0) {
+					message(MESS_ERROR,
+						"getting default context: %s\n",
+						strerror(errno));
+					if (selinux_enforce) {
+						freecon(oldContext);
+						if (close(fdcurr) < 0)
+							message(MESS_ERROR, "error closing file %s",
+									log->files[logNum]);
+						return 1;
+					}
+				}
+				if (setfscreatecon_raw(oldContext) < 0) {
+					message(MESS_ERROR,
+						"setting file context %s to %s: %s\n",
+						log->files[logNum], oldContext, strerror(errno));
+					if (selinux_enforce) {
+						freecon(oldContext);
+						if (close(fdcurr) < 0)
+							message(MESS_ERROR, "error closing file %s",
+									log->files[logNum]);
+						return 1;
+					}
+				}
+				message(MESS_DEBUG, "fscreate context set to %s\n",
+						oldContext);
+				freecon(oldContext);
+			} else {
+				if (errno != ENOTSUP) {
+					message(MESS_ERROR, "getting file context %s: %s\n",
+						log->files[logNum], strerror(errno));
+					if (selinux_enforce) {
+						if (close(fdcurr) < 0)
+							message(MESS_ERROR, "error closing file %s",
+									log->files[logNum]);
+						return 1;
+					}
+				}
+			}
+			if (close(fdcurr) < 0)
+				message(MESS_ERROR, "error closing file %s",
+						log->files[logNum]);
+		}
+#endif
+		message(MESS_DEBUG, "renaming %s to %s\n", log->files[logNum],
+		    rotNames->finalName);
 	    if (!debug && !hasErrors &&
 		rename(log->files[logNum], rotNames->finalName)) {
 		message(MESS_ERROR, "failed to rename %s to %s: %s\n",
@@ -903,19 +1109,26 @@ int rotateSingleLog(logInfo * log, int logNum, logState * state,
 	    else
 		sb.st_mode = log->createMode;
 
-	    message(MESS_DEBUG, "creating new log mode = 0%o uid = %d "
-		    "gid = %d\n", (unsigned int) sb.st_mode,
+	    message(MESS_DEBUG, "creating new %s mode = 0%o uid = %d "
+		    "gid = %d\n", log->files[logNum], (unsigned int) sb.st_mode,
 		    (int) sb.st_uid, (int) sb.st_gid);
 
 	    if (!debug) {
-		fd = createOutputFile(log->files[logNum], O_CREAT | O_RDWR,
-				      &sb);
-		if (fd < 0)
-		    hasErrors = 1;
-                else
-                    close(fd);
+			fd = createOutputFile(log->files[logNum], O_CREAT | O_RDWR,
+						  &sb);
+			if (fd < 0)
+				hasErrors = 1;
+			else
+				close(fd);
 	    }
 	}
+#ifdef WITH_SELINUX
+	if (selinux_enabled) {
+	    setfscreatecon_raw(savedContext);
+		freecon(savedContext);
+		savedContext = NULL;
+	}
+#endif
 
 	if (!hasErrors
 	    && log->flags & (LOG_FLAG_COPYTRUNCATE | LOG_FLAG_COPY))
@@ -927,8 +1140,8 @@ int rotateSingleLog(logInfo * log, int logNum, logState * state,
     return hasErrors;
 }
 
-int postrotateSingleLog(logInfo * log, int logNum, logState * state,
-			logNames * rotNames)
+int postrotateSingleLog(struct logInfo *log, int logNum, struct logState *state,
+			struct logNames *rotNames)
 {
     int hasErrors = 0;
 
@@ -954,28 +1167,26 @@ int postrotateSingleLog(logInfo * log, int logNum, logState * state,
     }
 
     if (!hasErrors && rotNames->disposeName)
-	hasErrors = removeLogFile(rotNames->disposeName);
+	hasErrors = removeLogFile(rotNames->disposeName, log);
 
 #ifdef WITH_SELINUX
-    if (selinux_enabled) {
-	setfscreatecon_raw(prev_context);
-	if (prev_context != NULL) {
-	    freecon(prev_context);
-	    prev_context = NULL;
+	if (selinux_enabled) {
+		setfscreatecon_raw(prev_context);
+		freecon(prev_context);
+		prev_context = NULL;
 	}
-    }
 #endif
     return hasErrors;
 }
 
-int rotateLogSet(logInfo * log, struct stateSet *sip, int force)
+int rotateLogSet(struct logInfo *log, int force)
 {
     int i, j;
     int hasErrors = 0;
     int logHasErrors[log->numFiles];
     int numRotated = 0;
-    logState **state;
-    logNames **rotNames;
+    struct logState **state;
+    struct logNames **rotNames;
 
     if (force)
 	log->criterium = ROT_FORCE;
@@ -1025,11 +1236,11 @@ int rotateLogSet(logInfo * log, struct stateSet *sip, int force)
     }
 
     for (i = 0; i < log->numFiles; i++) {
-	logHasErrors[i] = findNeedRotating(log, i, sip);
+	logHasErrors[i] = findNeedRotating(log, i);
 	hasErrors |= logHasErrors[i];
 
 	/* sure is a lot of findStating going on .. */
-	if ((findState(log->files[i], sip))->doRotate)
+	if ((findState(log->files[i]))->doRotate)
 	    numRotated++;
     }
 
@@ -1043,12 +1254,14 @@ int rotateLogSet(logInfo * log, struct stateSet *sip, int force)
 		message(MESS_ERROR, "error running first action script "
 			"for %s\n", log->pattern);
 		hasErrors = 1;
+		/* finish early, firstaction failed, affects all logs in set */
+		return hasErrors;
 	    }
 	}
     }
 
-    state = malloc(log->numFiles * sizeof(logState *));
-    rotNames = malloc(log->numFiles * sizeof(logNames *));
+    state = malloc(log->numFiles * sizeof(struct logState *));
+    rotNames = malloc(log->numFiles * sizeof(struct logNames *));
 
     for (j = 0;
 	 (!(log->flags & LOG_FLAG_SHAREDSCRIPTS) && j < log->numFiles)
@@ -1057,26 +1270,35 @@ int rotateLogSet(logInfo * log, struct stateSet *sip, int force)
 	for (i = j;
 	     ((log->flags & LOG_FLAG_SHAREDSCRIPTS) && i < log->numFiles)
 	     || (!(log->flags & LOG_FLAG_SHAREDSCRIPTS) && i == j); i++) {
-	    state[i] = findState(log->files[i], sip);
+	    state[i] = findState(log->files[i]);
 
-	    rotNames[i] = malloc(sizeof(logNames));
-	    memset(rotNames[i], 0, sizeof(logNames));
+	    rotNames[i] = malloc(sizeof(struct logNames));
+	    memset(rotNames[i], 0, sizeof(struct logNames));
 
 	    logHasErrors[i] |=
 		prerotateSingleLog(log, i, state[i], rotNames[i]);
 	    hasErrors |= logHasErrors[i];
 	}
 
-	if (log->pre) {
+	if (log->pre
+	    && (! ( (logHasErrors[j] && !(log->flags & LOG_FLAG_SHAREDSCRIPTS))
+		   || (hasErrors && (log->flags & LOG_FLAG_SHAREDSCRIPTS)) ) )) {
 	    if (!numRotated) {
 		message(MESS_DEBUG, "not running prerotate script, "
 			"since no logs will be rotated\n");
 	    } else {
 		message(MESS_DEBUG, "running prerotate script\n");
 		if (runScript(log->pattern, log->pre)) {
-		    message(MESS_ERROR,
-			    "error running shared prerotate script "
-			    "for %s\n", log->pattern);
+		    if (log->flags & LOG_FLAG_SHAREDSCRIPTS)
+			message(MESS_ERROR,
+				"error running shared prerotate script "
+				"for '%s'\n", log->pattern);
+		    else {
+			message(MESS_ERROR,
+				"error running non-shared prerotate script "
+				"for %s of '%s'\n", log->files[j], log->pattern);
+		    }
+		    logHasErrors[j] = 1;
 		    hasErrors = 1;
 		}
 	    }
@@ -1085,22 +1307,33 @@ int rotateLogSet(logInfo * log, struct stateSet *sip, int force)
 	for (i = j;
 	     ((log->flags & LOG_FLAG_SHAREDSCRIPTS) && i < log->numFiles)
 	     || (!(log->flags & LOG_FLAG_SHAREDSCRIPTS) && i == j); i++) {
-	    if (!logHasErrors[i]) {
+	    if (! ( (logHasErrors[i] && !(log->flags & LOG_FLAG_SHAREDSCRIPTS))
+		   || (hasErrors && (log->flags & LOG_FLAG_SHAREDSCRIPTS)) ) ) {
 		logHasErrors[i] |=
 		    rotateSingleLog(log, i, state[i], rotNames[i]);
 		hasErrors |= logHasErrors[i];
 	    }
 	}
 
-	if (log->post) {
+	if (log->post
+	    && (! ( (logHasErrors[j] && !(log->flags & LOG_FLAG_SHAREDSCRIPTS))
+		   || (hasErrors && (log->flags & LOG_FLAG_SHAREDSCRIPTS)) ) )) {
 	    if (!numRotated) {
 		message(MESS_DEBUG, "not running postrotate script, "
 			"since no logs were rotated\n");
 	    } else {
 		message(MESS_DEBUG, "running postrotate script\n");
 		if (runScript(log->pattern, log->post)) {
-		    message(MESS_ERROR, "error running postrotate script "
-			    "for %s\n", log->pattern);
+		    if (log->flags & LOG_FLAG_SHAREDSCRIPTS)
+			message(MESS_ERROR,
+				"error running shared postrotate script "
+				"for '%s'\n", log->pattern);
+		    else {
+			message(MESS_ERROR,
+				"error running non-shared postrotate script "
+				"for %s of '%s'\n", log->files[j], log->pattern);
+		    }
+		    logHasErrors[j] = 1;
 		    hasErrors = 1;
 		}
 	    }
@@ -1109,7 +1342,8 @@ int rotateLogSet(logInfo * log, struct stateSet *sip, int force)
 	for (i = j;
 	     ((log->flags & LOG_FLAG_SHAREDSCRIPTS) && i < log->numFiles)
 	     || (!(log->flags & LOG_FLAG_SHAREDSCRIPTS) && i == j); i++) {
-	    if (!logHasErrors[i]) {
+	    if (! ( (logHasErrors[i] && !(log->flags & LOG_FLAG_SHAREDSCRIPTS))
+		   || (hasErrors && (log->flags & LOG_FLAG_SHAREDSCRIPTS)) ) ) {
 		logHasErrors[i] |=
 		    postrotateSingleLog(log, i, state[i], rotNames[i]);
 		hasErrors |= logHasErrors[i];
@@ -1146,14 +1380,14 @@ int rotateLogSet(logInfo * log, struct stateSet *sip, int force)
     return hasErrors;
 }
 
-static int writeState(char *stateFilename, struct stateSet si)
+static int writeState(char *stateFilename)
 {
+    struct logState *p;
     FILE *f;
     char *chptr;
     int i;
 
     f = fopen(stateFilename, "w");
-
     if (!f) {
 	message(MESS_ERROR, "error creating state file %s: %s\n",
 		stateFilename, strerror(errno));
@@ -1162,30 +1396,32 @@ static int writeState(char *stateFilename, struct stateSet si)
 
     fprintf(f, "logrotate state -- version 2\n");
 
-    for (i = 0; i < si.numStates; i++) {
-	fputc('"', f);
-	for (chptr = si.states[i].fn; *chptr; chptr++) {
-	    switch (*chptr) {
-	    case '"':
-		fputc('\\', f);
-	    }
+	for (i = 0; i < hashSize; i++) {
+		for (p = states[i]->head.lh_first; p != NULL;
+				p = p->list.le_next) {
+			fputc('"', f);
+			for (chptr = p->fn; *chptr; chptr++) {
+				switch (*chptr) {
+				case '"':
+					fputc('\\', f);
+				}
 
-	    fputc(*chptr, f);
+				fputc(*chptr, f);
+			}
+
+			fputc('"', f);
+			fprintf(f, " %d-%d-%d\n",
+			p->lastRotated.tm_year + 1900,
+			p->lastRotated.tm_mon + 1,
+			p->lastRotated.tm_mday);
+		}
 	}
 
-	fputc('"', f);
-	fprintf(f, " %d-%d-%d\n",
-		si.states[i].lastRotated.tm_year + 1900,
-		si.states[i].lastRotated.tm_mon + 1,
-		si.states[i].lastRotated.tm_mday);
-    }
-
-    fclose(f);
-
-    return 0;
+	fclose(f);
+	return 0;
 }
 
-static int readState(char *stateFilename, struct stateSet *sip)
+static int readState(char *stateFilename)
 {
     FILE *f;
     char buf[1024];
@@ -1195,7 +1431,7 @@ static int readState(char *stateFilename, struct stateSet *sip)
     int i;
     int line = 0;
     int error;
-    logState *st;
+    struct logState *st;
     time_t lr_time;
     struct stat f_stat;
 
@@ -1244,6 +1480,7 @@ static int readState(char *stateFilename, struct stateSet *sip)
     line++;
 
     while (fgets(buf, sizeof(buf) - 1, f)) {
+	argv = NULL;
 	line++;
 	i = strlen(buf);
 	if (buf[i - 1] != '\n') {
@@ -1262,7 +1499,6 @@ static int readState(char *stateFilename, struct stateSet *sip)
 	    (sscanf(argv[1], "%d-%d-%d", &year, &month, &day) != 3)) {
 	    message(MESS_ERROR, "bad line %d in state file %s\n",
 		    line, stateFilename);
-	    if (argv)
 		free(argv);
 	    fclose(f);
 	    return 1;
@@ -1299,7 +1535,8 @@ static int readState(char *stateFilename, struct stateSet *sip)
 
 	year -= 1900, month -= 1;
 
-	st = findState(argv[0], sip);
+	if ((st = findState(argv[0])) == NULL)
+		return 1;
 
 	st->lastRotated.tm_mon = month;
 	st->lastRotated.tm_mday = day;
@@ -1313,23 +1550,20 @@ static int readState(char *stateFilename, struct stateSet *sip)
     }
 
     fclose(f);
-
     return 0;
-
 }
 
 int main(int argc, const char **argv)
 {
-    int numLogs = 0;
     int force = 0;
-    logInfo *logs = NULL;
-    struct stateSet si = { NULL, 0 };
     char *stateFile = STATEFILE;
-    int i;
     int rc = 0;
     int arg;
     const char **files;
     poptContext optCon;
+	struct logInfo *log;
+	int state_file_ok = 1;
+
     struct poptOption options[] = {
 	{"debug", 'd', 0, 0, 'd',
 	 "Don't do anything, just test (implies -v)"},
@@ -1386,33 +1620,39 @@ int main(int argc, const char **argv)
     selinux_enforce = security_getenforce();
 #endif
 
-    if (readAllConfigPaths(files, &logs, &numLogs)) {
+	TAILQ_INIT(&logs);
+
+	if (readAllConfigPaths(files)) {
 	poptFreeContext(optCon);
-	free_logInfo(&logs, &numLogs);
 	exit(1);
     }
 
     poptFreeContext(optCon);
     nowSecs = time(NULL);
 
-    if (readState(stateFile, &si)) {
-	exit(1);
-    }
+	if (allocateHash() != 0)
+		return 1;
 
-    message(MESS_DEBUG, "\nHandling %d logs\n", numLogs);
+	if (readState(stateFile))
+	{
+		state_file_ok = 0;
+		/* exit(1); */
+	}
 
-    for (i = 0; i < numLogs; i++) {
-	rc |= rotateLogSet(logs + i, &si, force);
-    }
+	message(MESS_DEBUG, "\nHandling %d logs\n", numLogs);
 
-    if (!debug)
-	rc |= writeState(stateFile, si);
+	for (log = logs.tqh_first; log != NULL; log = log->list.tqe_next)
+		rc |= rotateLogSet(log, force);
 
-    for (i = 0; i < si.numStates; i++) {
-	free(si.states[i].fn);
-    }
-    free(si.states);
+	if (!debug && state_file_ok)
+		rc |= writeState(stateFile);
 
-    free_logInfo(&logs, &numLogs);
-    return (rc != 0);
+	if (!state_file_ok)
+	{
+		message(MESS_ERROR, "could not read state file, "
+				"will not attempt to write into it\n");
+		rc = 1;
+	}
+	
+	return (rc != 0);
 }
