@@ -2,6 +2,7 @@
 #define H_LOGROTATE
 
 #include <sys/types.h>
+#include <sys/queue.h>
 #include <glob.h>
 
 #include "config.h"
@@ -16,6 +17,7 @@
 #define LOG_FLAG_SHAREDSCRIPTS	(1 << 7)
 #define LOG_FLAG_COPY		(1 << 8)
 #define LOG_FLAG_DATEEXT	(1 << 9)
+#define LOG_FLAG_SHRED		(1 << 10)
 
 #define NO_MODE ((mode_t) -1)
 #define NO_UID  ((uid_t) -1)
@@ -24,8 +26,7 @@
 #define NO_FORCE_ROTATE 0
 #define FORCE_ROTATE    1
 
-
-typedef struct {
+struct logInfo {
     char *pattern;
     char **files;
     int numFiles;
@@ -43,18 +44,23 @@ typedef struct {
     char *compress_prog;
     char *uncompress_prog;
     char *compress_ext;
+	char *dateformat;		/* specify format for strftime (for dateext) */
     int flags;
+	int shred_cycles;		/* if !=0, pass -n shred_cycles to GNU shred */
     mode_t createMode;		/* if any/all of these are -1, we use the */
     uid_t createUid;		/* attributes from the log file just rotated */
     gid_t createGid;
     /* these are at the end so they end up nil */
     const char **compress_options_list;
     int compress_options_count;
-} logInfo;
+    TAILQ_ENTRY(logInfo) list;
+};
 
-int readAllConfigPaths(const char **paths, logInfo ** logsPtr,
-		       int *numLogsPtr);
+TAILQ_HEAD(logInfoHead, logInfo) logs;
 
+extern int numLogs;
 extern int debug;
 
+int readAllConfigPaths(const char **paths);
+
 #endif
